from .otBase import BaseTTXConverter


class table__f_e_a_t(BaseTTXConverter):
	"""The feature name table is an AAT (Apple Advanced Typography) table for
	storing font features, settings, and their human-readable names. It should
	not be confused with the ``Feat`` table or the OpenType Layout ``GSUB``/``GPOS``
	tables. See `Feature Name Table <https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6feat.html>`_
	in the TrueType Reference Manual for more information on the structure and
	purpose of this table."""
	pass
